"""
Create or import top-level classes/methods.
"""
from typing import Tuple, Union
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
    from statsmodels.regression.linear_model import (
        RegressionResultsWrapper as StatsModel
    )


# Here we list any class we might use. In principle, they should be expected to have a fit() and a predict(X)
# method.
# However, the statsmodels library provides Model and Results classes. The former only hold X and y data, while the
# latter represents a fitted (parameterised) Model (with a working predict method), plus the result statistics
# (the Model object is also in there). We will for the foreseeable future model from statsmodels, but if we add
# others there might be a small refactoring necessary.
MODEL_CLASSES = (StatsModel,)
MODEL_TYPES = Union[MODEL_CLASSES]


# First public import block
from ts_forecasting_pipeline.speccing import (
    ModelSpecs,
    ObjectSeriesSpecs,
    DFFileSeriesSpecs,
    DBSeriesSpecs,
)


class ModelState(object):
    """
    This class abstracts all information we need to describe a model and how it was made.
    Container to hold all model-relevant state information:
    The trained model and the ModelsSpecs it is based on."""

    model: MODEL_TYPES  # a fitted model
    specs: ModelSpecs

    def __init__(self, model, specs):
        if not isinstance(model, MODEL_CLASSES):
            raise Exception(
                "ModelState(): model parameter needs to be of type %s"
                % str(MODEL_CLASSES)
            )
        self.model = model
        if not isinstance(specs, ModelSpecs):
            raise Exception(
                "ModelState(): specs parameter needs to be of type <ModelSpecs>"
            )
        self.specs = specs

    def split(self) -> Tuple[MODEL_TYPES, ModelSpecs]:
        """Return model and specs separately"""
        return self.model, self.specs

    def __repr__(self):
        return "ModelState: <%s, %s>" % (self.model, self.specs)


# second public import block
from ts_forecasting_pipeline.modelling import (
    create_fitted_model,
    evaluate_models,
    model_param_grid_search,
)
from ts_forecasting_pipeline.forecasting import (
    make_forecast_for,
    make_rolling_forecasts,
)
from ts_forecasting_pipeline.serializing import save_model, load_model



from typing import Dict, Tuple, Optional, Sequence
from datetime import datetime

import pandas as pd
import numpy as np
import statsmodels.api as sm
from matplotlib import pyplot as plt

from ts_forecasting_pipeline.speccing import ModelSpecs
from ts_forecasting_pipeline.featuring import construct_features
from ts_forecasting_pipeline.serializing import save_model
from ts_forecasting_pipeline import MODEL_TYPES, ModelState


"""
Functions for working with time series models.
"""


def create_fitted_model(
    specs: ModelSpecs, version: str, save: bool = False
) -> MODEL_TYPES:
    """
    Create a new fitted model with the given specs.
    If needed, the fitted model and specs can be persisted to file. From there, they can be loaded back with
    load_model.

    TODO: frequency of 15T is still hardcoded.
    """
    regression_frame = construct_features(purpose="train", specs=specs)

    # Remove any observation where data is missing.
    # Other parts of the workflow cannot handle missing data, so everything should be verified here.
    regression_frame = regression_frame[~pd.isnull(regression_frame).any(axis=1)]

    x_train = regression_frame.iloc[:, 1:]
    y_train = np.array(regression_frame.iloc[:, 0])

    if specs.model_type == "OLS":
        model = sm.OLS(y_train, x_train)
    else:
        raise Exception("Unknown model type: %s " % specs.model_type)

    fitted_model = model.fit()

    if save:
        save_model(ModelState(fitted_model, specs), version)

    return fitted_model


def evaluate_models(
    m1: ModelState, m2: Optional[ModelState] = None, plot_path: str = None
):
    """
    Run a model or two against test data and plot results.
    Useful to judge model performance or compare two models.
    Shows RMSE values, plots error distributions and prints the time it took to forecast.

    TODO: support testing m2 next to m1
    """
    fitted_m1, m1_specs = m1.split()

    regression_frame = construct_features(purpose="test", specs=m1_specs)

    x_test = regression_frame.iloc[:, 1:]
    y_test = np.array(regression_frame.iloc[:, 0])

    y_hat_test = fitted_m1.predict(x_test)
    print(
        "rmse = %s"
        % (str(round(sm.tools.eval_measures.rmse(y_test, y_hat_test, axis=0), 4)))
    )

    plot_true_versus_predicted(
        regression_frame.index, y_test, y_hat_test.values, plot_path
    )

    plot_error_graph(y_test, y_hat_test, plot_path=plot_path)


def plot_true_versus_predicted(
    indices: pd.DatetimeIndex,
    true_values: Sequence[float],
    predicted_values: Sequence[float],
    plot_path: str = None,
):
    """
    Helper function to plot real values next to forecasted values.
    """
    plt.plot(indices.to_pydatetime(), true_values, label="actual")
    plt.plot(indices.to_pydatetime(), predicted_values, label="predicted")
    plt.xlabel("tested time steps")
    plt.ylabel("values")
    plt.legend()
    if plot_path is None:
        plt.show()
    else:
        plt.savefig(plot_path + "/true_versus_predicted.png")
        plt.close()


def plot_error_graph(
    true_values: Sequence[float],
    predicted_values: Sequence[float],
    use_abs_errors: bool = False,
    plot_path: str = None,
):

    results_df = pd.DataFrame({"y_hat_test": predicted_values, "y_test": true_values})

    # remove 0 s
    results_df = results_df[(results_df != 0).all(1)]

    results_df["max_error"] = abs(results_df.y_hat_test / results_df.y_test - 1)
    if use_abs_errors is True:
        #  if you want to look at abs values, instead of (abs)proportional errors
        results_df["max_error"] = abs(results_df.y_hat_test - results_df.y_test)

    results_df.sort_values("max_error", inplace=True)
    results_df["proportion"] = (np.arange(len(results_df)) + 1) / len(results_df)

    plt.plot(results_df["max_error"], results_df["proportion"], "-o")
    plt.ylim(0, 1)
    plt.xlim(0, 1)
    plt.xlabel("max error for proportion")
    plt.ylabel("proportion of observations")
    if plot_path is None:
        plt.show()
    else:
        plt.savefig(plot_path + "/error_graph.png")
        plt.close()


def model_param_grid_search(
    df: pd.DataFrame,
    start_of_training_data: datetime,
    end_of_test_data: datetime,
    params: Dict[str, Tuple[float, float]],
) -> Dict[str, float]:

    """
    Creates and tests models with different model parameters.
    Returns the best parameter set w.r.t. smallest RMSE.
    """
    return {}



"""
functionality to create the feature data necessary.
"""
from typing import List

import pandas as pd


def construct_dataframe(variable_containers):
    """
    variable_containers: list of tuples
    each tuple should contain : df, name of column in df, new name of column in resultin df
    data frame is created based on the index of input
    """
    dict_for_df = {}
    for variable_container in variable_containers:
        new_name = variable_container[2]
        cur_name = variable_container[1]
        df_data = variable_container[0]
        dict_for_df[new_name] = df_data[cur_name]
    df = pd.DataFrame(dict_for_df)
    return df


def convert_lag_to_name_str(lag):
    if lag < 0:
        str_lag = "l" + str(abs(lag))
    else:
        str_lag = "f" + str(abs(lag))
    return str_lag


def add_lags(df: pd.DataFrame, column: str, lags: List[int], inplace=True):
    """
    df: DataFrame 
    Contains data with the contains columns to lag

    column: str
    The column with the data to lag

    lags: list
    lags in days. negative means a lag, positive is a future value

    inplace: Bool


    Creates lag columns for a column in the dataframe (inplace)
    Negative values are lags, while positive values are future values
    The new columns are named like the lagged column, plus "_l<lag>".
    In case of positive 'lags' (future values), new columns are named like the lagged column, plus "_f<lag>".

    TODO: hardcoded the time interval (15 minutes), in case of other interval or missing rows, it does not return right value
    """
    if inplace == False:
        df = df.copy()

    for lag in lags:
        str_lag = convert_lag_to_name_str(lag)

        df[column + "_" + str_lag] = df[column].shift(-96 * lag)

    if inplace == False:
        return df


### note hendrik: don't fully understand this function. I created construct_dataframe(), does that replace this function?
def add_regressors(df: pd.DataFrame, name: str, horizons: List[str]):
    """
    Get regressor data to be added to df, in form of forecasted data.
    """
    # Calls forecasting.get_forecasts() on a specific pickle or table
    pass



from typing import List, Optional, Callable, Union
from datetime import datetime, timedelta, tzinfo
from pprint import pformat
import json

import pytz
import dateutil.parser
import pandas as pd
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Query

from ts_forecasting_pipeline.utils import tz_aware_utc_now


DEFAULT_RATIO_TRAINING_TEST_DATA = 2 / 3
DEFAULT_REMODELING_FREQUENCY = timedelta(days=1)


class SeriesSpecs(object):
    """Describes a time series (e.g. a pandas Series).
    In essence, a column in the regression frame, filled with numbers.

    Using this class, the column will be filled with NaN values.

    If you have data to be loaded in automatically, you should be using one of the subclasses, which allow to describe
    or pass in an actual data source to be loaded.

    When dealing with columns, our code should usually refer to this superclass so it does not need to care
    which kind of data source it is dealing with.
    """

    # The name in the resulting feature frame, and possibly in the saved model specs (named by outcome var)
    name: str
    # The name in the data source, if source is a pandas DataFrame or database Table - if None, the name will be tried
    column: Optional[str]
    # timezone of the data - useful when de-serializing (e.g. pandas serialises to UTC)
    original_tz: tzinfo

    def __init__(self, name: str, column: str = None, original_tz: tzinfo = None):
        self.name = name
        self.column = column
        if self.column is None:
            self.column = name
        self.original_tz = original_tz
        self.__series_type__ = self.__class__.__name__

    def as_dict(self):
        return vars(self)

    def load_series(self) -> pd.Series:
        return pd.Series()

    def __repr__(self):
        return "%s: <%s>" % (self.__class__.__name__, self.as_dict())


class ObjectSeriesSpecs(SeriesSpecs):
    """
    Spec for a pd.Series object that is being passed in and is stored directly in the specs.
    The data is not mutatable after creation.
    Note: The column argument is not used, as the series has only one column.
    """

    data: pd.Series

    def __init__(
        self, data: pd.Series, name: str, column: str = None, original_tz: tzinfo = None
    ):
        super().__init__(name, None, original_tz)
        self.data = data
        self.original_tz = data.index.tzinfo
        if self.original_tz is None:
            self.original_tz = pytz.utc

    def load_series(self) -> pd.Series:
        return self.data


class DFFileSeriesSpecs(SeriesSpecs):
    """
    Spec for a pandas DataFrame source.
    This class holds the filename, from which we unpickle the data frame, then read the column.
    """

    file_path: str

    def __init__(
        self, file_path: str, name: str, column: str = None, original_tz: tzinfo = None
    ):
        super().__init__(name, column, original_tz)
        self.file_path = file_path

    def load_series(self) -> pd.Series:
        df: pd.DataFrame = pd.read_pickle(self.file_path)
        if df.index.tzinfo is None:
            self.original_tz = pytz.utc
            df.index = df.index.tz_localize(self.original_tz)
        else:
            self.original_tz = df.index.tzinfo
        return df[self.column]


class DBSeriesSpecs(SeriesSpecs):

    """Define how to query a database for time series values.
    This works via a SQLAlchemy query.
    This query should return the needed information for the forecasting pipeline:
    A "datetime" column (which will be set as index) and the values column (named by name or column,
    see SeriesSpecs.__init__, defaults to "value"). For example:
    TODO: show an example"""

    db: Engine
    query: Query

    def __init__(
        self,
        db_engine: Engine,
        query: Query,
        name: str = "value",
        column: str = None,
        original_tz: tzinfo = pytz.utc,  # postgres stores naive datetimes
    ):
        super().__init__(name, column, original_tz)
        self.db_engine = db_engine
        self.query = query

    def load_series(self) -> pd.Series:
        series_orig = pd.read_sql(
            self.query.statement, self.db_engine, parse_dates=["datetime"]
        )
        series_orig.set_index("datetime", drop=True, inplace=True)
        if series_orig.index.tzinfo is None:
            if self.original_tz is not None:
                series_orig.index = series_orig.index.tz_localize(self.original_tz)
        else:
            series_orig.index = series_orig.index.tz_convert(self.original_tz)
        series = series_orig.resample("15T").mean()
        return series["value"]


transform_type = Optional[Callable[[pd.DataFrame], pd.DataFrame]]


class ModelSpecs(object):
    """Describes a model and how it was trained.
    """

    outcome_var: SeriesSpecs
    model_type: str  # e.g. OLS, ARIMA, ...
    lags: List[int]
    regressors: List[SeriesSpecs]
    # Start of training data set
    start_of_training_data: datetime
    # End of testing data set
    end_of_test_data: datetime
    # This determines the cutoff point between training and testing data
    ratio_training_test_data: float
    # time this model was created, defaults to UTC now
    creation_time: datetime
    valid_until: datetime
    model_filename: str
    remodel_frequency: timedelta
    # Custom transformation to perform on the data. Called after all SeriesSpecs were resolved.
    transform: transform_type

    def __init__(
        self,
        outcome_var: Union[str, SeriesSpecs, pd.Series],
        model_type: str,
        start_of_training_data: Union[str, datetime],
        end_of_test_data: Union[str, datetime],
        lags: List[int] = None,
        regressors: Union[List[str], List[SeriesSpecs], List[pd.Series]] = None,
        ratio_training_test_data=DEFAULT_RATIO_TRAINING_TEST_DATA,
        remodel_frequency: Union[str, timedelta] = DEFAULT_REMODELING_FREQUENCY,
        model_filename: str = None,
        creation_time: Union[str, datetime] = None,
        transform: transform_type = None,
    ):
        """Create a ModelSpecs instance. Accepts all parameters as string (besides transform - TODO) for
         deserialization support (JSON strings for all parameters which are not natively JSON-parseable,)"""
        self.outcome_var = parse_series_specs(outcome_var, "y")
        self.model_type = model_type
        self.lags = lags
        if self.lags is None:
            self.lags = []
        if regressors is None:
            self.regressors = []
        else:
            self.regressors = [
                parse_series_specs(r, "Regressor%d" % (regressors.index(r) + 1))
                for r in regressors
            ]
        if isinstance(start_of_training_data, str):
            self.start_of_training_data = dateutil.parser.parse(start_of_training_data)
        else:
            self.start_of_training_data = start_of_training_data
        if isinstance(end_of_test_data, str):
            self.end_of_test_data = dateutil.parser.parse(end_of_test_data)
        else:
            self.end_of_test_data = end_of_test_data
        self.ratio_training_test_data = ratio_training_test_data
        if isinstance(creation_time, str):
            self.creation_time = dateutil.parser.parse(creation_time)
        elif creation_time is None:
            self.creation_time = tz_aware_utc_now()
        else:
            self.creation_time = creation_time
        self.model_filename = model_filename
        if isinstance(remodel_frequency, str):
            self.remodel_frequency = timedelta(
                days=int(remodel_frequency) / 60 / 60 / 24
            )
        else:
            self.remodel_frequency = remodel_frequency
        self.transform = transform
        # TODO: load transform function. Not sure what is smart, maybe marshalling.

    def as_dict(self):
        return vars(self)

    def __repr__(self):
        return "ModelSpecs: <%s>" % pformat(vars(self))


def parse_series_specs(
    specs: Union[str, SeriesSpecs, pd.Series], name: str = None
) -> SeriesSpecs:
    if isinstance(specs, str):
        return load_series_specs_from_json(specs)
    elif isinstance(specs, pd.Series):
        return ObjectSeriesSpecs(specs, name)
    else:
        return specs


def load_series_specs_from_json(s: str) -> SeriesSpecs:
    json_repr = json.loads(s)
    series_class = globals()[json_repr["__series_type__"]]
    if series_class == ObjectSeriesSpecs:
        # load pd.Series from string, will be UTC-indexed, so apply original_tz
        json_repr["data"] = pd.read_json(
            json_repr["data"], typ="series", convert_dates=True
        )
        json_repr["data"].index = json_repr["data"].index.tz_localize(
            json_repr["original_tz"]
        )
    return series_class(
        **{k: v for k, v in json_repr.items() if not k.startswith("__")}
    )
